package ch.SWITCH.aai.uApprove.components;

import java.io.Serializable;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.ArrayList;

import org.json.simple.JSONObject;
import org.json.simple.JSONValue;
import org.json.simple.JSONArray;
import org.opensaml.saml2.metadata.AttributeConsumingService;
import org.opensaml.saml2.metadata.LocalizedString;
import org.opensaml.saml2.metadata.ServiceDescription;
import org.opensaml.saml2.metadata.ServiceName;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.opensaml.saml2.metadata.RequestedAttribute;
import org.opensaml.xml.util.AttributeMap;

import javax.xml.namespace.QName;

import jp.gakunin.uApprovejp.saml2.metadata.Description;
import jp.gakunin.uApprovejp.saml2.metadata.RequestedAttributeExtension;
import jp.gakunin.uApprovejp.saml2.metadata.UApprovejpMetadata;

public class RelyingParty implements Serializable {

	private static final long serialVersionUID = -2802281149448775031L;
	private static final QName ATTRDESC_NAME = new QName(UApprovejpMetadata.NAMESPACE, "description");

	private static final String RA_DESCRIPTIONS = "raDescriptions";
	private static final String KEY_DESCRIPTION = "description";
	private static final String KEY_LOCALIZED_DM = "localizedDescMap";

	private final Logger logger = LoggerFactory.getLogger(RelyingParty.class);

	private String entityId;
	private Map<String, String> rpNames;
	private Map<String, String> rpDescriptions;
	private List<String> mandatoryAttributes;
	private Map<String, RequestedAttributeDescription> raDescriptions;

	/**
	 * Returns a new RelyingParty object.
	 * 
	 * @param entityId the entity ID of the relying party
	 * @param attrService SAML 2.0 Metadata AttributeConsumingService
	 * @param raExtensions 
	 * @return a new RelyingParty object.
	 */
	public static RelyingParty newInstance(String entityId,
			AttributeConsumingService attrService, List<RequestedAttributeExtension> raExtensions) {
		RelyingParty relyingParty = new RelyingParty(entityId, attrService, raExtensions);
		return relyingParty;
	}

	/**
	 * Returns a new RelyingParty object.
	 * 
	 * @param serialized
	 * @return a new RelyingParty object.
	 */
	public static RelyingParty newInstance(String serialized) {
		JSONObject rpSerialized = (JSONObject) JSONValue.parse(serialized);
		String entityId = (String) rpSerialized.get("entityId");
		RelyingParty relyingParty = new RelyingParty(entityId, rpSerialized);
		return relyingParty;
	}

	private RelyingParty(String entityId) {
		this.entityId = entityId;

		rpNames = new HashMap<String, String>();
		rpDescriptions = new HashMap<String, String>();
		mandatoryAttributes = new ArrayList<String>();
		raDescriptions = new HashMap<String, RequestedAttributeDescription>();
	}

	private RelyingParty(String entityId, AttributeConsumingService attrService, List<RequestedAttributeExtension> raExtensions) {
		this(entityId);

		logger.debug("creating RelyingParty instance.");
		if (attrService != null) {
			setRpNames(attrService.getNames());
			setRpDescriptions(attrService.getDescriptions());
			setRequestedAttributes(attrService.getRequestAttributes());
		}

		if (raExtensions != null) {
			Map<String, List< Description> > map = new HashMap<String, List< Description > >();
			for (RequestedAttributeExtension raExtension: raExtensions) {
				String friendlyName = raExtension.getFriendlyName();
				logger.debug("RequestedAttributeExtension.FriendryName=" + friendlyName);

				for (Description desc: raExtension.getDescriptions()) {
					LocalizedString description = desc.getDescription();
					logger.debug("Description.lang=" + description.getLanguage());
					logger.debug("Description.desc=" + description.getLocalString());
				}

				map.put(friendlyName, raExtension.getDescriptions());
			}
			setRequestedAttributeExtensions(attrService.getRequestAttributes(),	map);
		}
	}

	private RelyingParty(String entityId, JSONObject rpSerialized) {
		this(entityId);
	
		setRpNames(rpSerialized);
		setRpDescriptions(rpSerialized);
		setRequestedAttributes(rpSerialized);
		setRequestedAttributeExtensions(rpSerialized);
	}

	private void setRpNames(List<ServiceName> names) {
		for (ServiceName element : names) {
			rpNames.put(element.getName().getLanguage(), element.getName().getLocalString());
		}
	}

	private void setRpNames(JSONObject rpSerialized) {
		JSONObject rpNamesSerialized = (JSONObject) rpSerialized.get("rpNames");
		for (Object key : rpNamesSerialized.keySet()) {
			rpNames.put((String) key, (String) rpNamesSerialized.get(key));
		}
	}

	private void setRpDescriptions(List<ServiceDescription> descriptions) {
		for (ServiceDescription element : descriptions) {
			rpDescriptions.put(element.getDescription().getLanguage(), element.getDescription().getLocalString());
		}
	}

	private void setRpDescriptions(JSONObject rpSerialized) {
		JSONObject rpDescriptionsSerialized = (JSONObject) rpSerialized.get("rpDescriptions");
		for (Object key : rpDescriptionsSerialized.keySet()) {
			rpDescriptions.put((String) key, (String) rpDescriptionsSerialized.get(key));
		}
	}

	private void setRequestedAttributes(List<RequestedAttribute> attributes) {
		for (RequestedAttribute element: attributes) {
			String friendlyName = element.getFriendlyName();

			if (element.isRequired()) {
				mandatoryAttributes.add(friendlyName);
			}

			if (logger.isDebugEnabled()) {
				AttributeMap map = element.getUnknownAttributes();
				logger.debug("{}: map key:", friendlyName);
				for (QName name : map.keySet()) {
					logger.debug(name.toString());
				}
			}

			String attrDesc = element.getUnknownAttributes().get(ATTRDESC_NAME);
			if (attrDesc != null) {
				if (!raDescriptions.containsKey(friendlyName)) {
					RequestedAttributeDescription raDesc = new RequestedAttributeDescription();
					raDesc.setDescription(attrDesc);
					raDescriptions.put(friendlyName, raDesc);
				}
			}
		}
	}
	
	private void setRequestedAttributes(JSONObject rpSerialized) {
		JSONArray mandatoryAttributesSerialized = (JSONArray) rpSerialized.get("mandatoryAttributes");
		for (int i = 0; i < mandatoryAttributesSerialized.size(); i++) {
			mandatoryAttributes.add((String) mandatoryAttributesSerialized.get(i));
		}
	}

	private void setRequestedAttributeExtensions(
			List<RequestedAttribute> attributes, Map<String, List<Description>> map) {
		for (RequestedAttribute element : attributes) {
			RequestedAttributeDescription raDesc = null;

			String friendlyName = element.getFriendlyName();
			if (!raDescriptions.containsKey(friendlyName)) {
				raDesc = new RequestedAttributeDescription();
				raDescriptions.put(friendlyName, raDesc);
			}

			if (map.containsKey(friendlyName)) {
				raDesc = raDescriptions.get(friendlyName);

				List<Description> list = map.get(friendlyName);
				for (Description desc : list) {
					LocalizedString description = desc.getDescription();
					raDesc.putLocalizedDescription(description);
				}
			}
		}
	}

	private void setRequestedAttributeExtensions(JSONObject rpSerialized) {
		JSONObject raDescriptionsSerialized = (JSONObject) rpSerialized
				.get(RA_DESCRIPTIONS);
		for (Object friendlyName : raDescriptionsSerialized.keySet()) {
			JSONObject raDescSerialized = (JSONObject) raDescriptionsSerialized
					.get(friendlyName);

			RequestedAttributeDescription raDesc = new RequestedAttributeDescription();

			String description = (String) raDescSerialized.get(KEY_DESCRIPTION);
			raDesc.setDescription(description);

			JSONObject localizedDescMapSerialized = (JSONObject) raDescSerialized
					.get(KEY_LOCALIZED_DM);
			Map<String, String> localizedDescMap = raDesc.getLocalizedDescMap();
			for (Object lang : localizedDescMapSerialized.keySet()) {
				localizedDescMap.put((String) lang,
						(String) localizedDescMapSerialized.get(lang));
			}

			raDescriptions.put((String) friendlyName, raDesc);
		}
	}

	@SuppressWarnings("unchecked")
	public String serialize() {
		logger.trace("Attempt to serialize data for {}", entityId);
		JSONObject result = new JSONObject();
		
		logger.trace("Attempt to serialize names {}", rpNames);
	    JSONObject rpNamesSerialized = new JSONObject();
	    for (String key : rpNames.keySet()) {
	    	rpNamesSerialized.put(key, rpNames.get(key));
	    }	    
		logger.trace("Serialized names {}", rpNamesSerialized);
	    
		logger.trace("Attempt to serialize description {}", rpDescriptions);
	    JSONObject rpDescriptionsSerialized = new JSONObject();
	    for (String key : rpDescriptions.keySet()) {
	    	rpDescriptionsSerialized.put(key, rpDescriptions.get(key));
	    }    
		logger.trace("Serialized descriptions {}", rpDescriptionsSerialized);

		JSONArray mandatoryAttributesSerialized = new JSONArray();
		for (int i = 0; i < mandatoryAttributes.size(); i++) {
			mandatoryAttributesSerialized.add(mandatoryAttributes.get(i));
		}

		JSONObject raDescriptionsSerialized = new JSONObject();
		for (String key : raDescriptions.keySet()) {
			RequestedAttributeDescription value = raDescriptions.get(key);
			String description = value.getDescription();

			JSONObject localizedDescMapSerialized = new JSONObject();
			for (String lang : value.keySet()) {
				String desc = value.getLocalizedDescription(lang);
				localizedDescMapSerialized.put(lang, desc);
			}
			JSONObject raDescSerialized = new JSONObject();
			raDescSerialized.put(KEY_DESCRIPTION, description);
			raDescSerialized.put(KEY_LOCALIZED_DM, localizedDescMapSerialized);
			raDescriptionsSerialized.put(key, raDescSerialized);
		}
		
		result.put("entityId", entityId);
		result.put("rpNames", rpNamesSerialized);
		result.put("rpDescriptions", rpDescriptionsSerialized);
		result.put("mandatoryAttributes", mandatoryAttributesSerialized);
		result.put(RA_DESCRIPTIONS, raDescriptionsSerialized);
		
		logger.trace("Serialized data '{}'", result);
		return result.toString();
	}

	public String getEntityId() {
		return entityId;
	}

	public Map<String, String> getRpNames() {
		return rpNames;
	}

	public Map<String, String> getRpDescriptions() {
		return rpDescriptions;
	}
	
	public List<String> getMandatoryAttributes() {
		return mandatoryAttributes;
	}
 
	public Map<String, RequestedAttributeDescription> getReqdAttrDescs() {
		return raDescriptions;
	}
	

}
