package ch.SWITCH.aai.uApprove.idpplugin;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import jp.gakunin.uApprovejp.saml2.metadata.Description;
import jp.gakunin.uApprovejp.saml2.metadata.RequestedAttributeExtension;
import jp.gakunin.uApprovejp.saml2.metadata.impl.DescriptionBuilder;
import jp.gakunin.uApprovejp.saml2.metadata.impl.RequestedAttributeExtensionBuilder;

import org.opensaml.common.xml.SAMLConstants;
import org.opensaml.saml2.common.Extensions;
import org.opensaml.saml2.metadata.AttributeConsumingService;
import org.opensaml.saml2.metadata.EntityDescriptor;
import org.opensaml.saml2.metadata.LocalizedString;
import org.opensaml.saml2.metadata.SPSSODescriptor;
import org.opensaml.saml2.metadata.provider.MetadataProvider;
import org.opensaml.saml2.metadata.provider.MetadataProviderException;
import org.opensaml.xml.XMLObject;
import org.opensaml.xml.schema.XSAny;
import org.opensaml.xml.util.AttributeMap;
import org.opensaml.xml.util.XMLConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.SWITCH.aai.uApprove.components.RelyingParty;
import ch.SWITCH.aai.uApprove.components.UApproveException;


public class MetadataAccess {
	private static Logger LOG = LoggerFactory.getLogger(MetadataAccess.class);
	private MetadataProvider metadataProvider;
	private static final QName FRIENDORY_NAME = new QName(RequestedAttributeExtension.FRIENDLY_NAME_ATTRIB_NAME);
	private static final QName XML_LANG = new QName(XMLConstants.XML_NS, Description.LANG_ATTRIB_NAME, XMLConstants.XML_PREFIX);
	private static final String[] PROTOCOLS = {SAMLConstants.SAML20P_NS, SAMLConstants.SAML11P_NS, SAMLConstants.SAML10P_NS};
	
	public MetadataAccess(MetadataProvider metadataProvider) {
	  this.metadataProvider = metadataProvider;
	}
	
	public RelyingParty getRelyingPartyInfo(String entityId, EntityDescriptor entityDescriptor) throws UApproveException {
		LOG.debug("RelyingParty info for entityId {}", entityId);
		
		RelyingParty relyingParty = buildRelyingParty(entityId, entityDescriptor);
		return relyingParty;
	}

	/**
	 * create RelyingParty object from
	 * /EntitiesDescriptor/EntityDescriptor/SPSSODescriptor element.
	 */
	private RelyingParty buildRelyingParty(String entityId,
			EntityDescriptor entityDescriptor) {
		AttributeConsumingService attrServ = null;
		List<RequestedAttributeExtension> raExtensions = null;

		SPSSODescriptor spSSODescriptor = getSPSSODescriptor(entityDescriptor);

		if (spSSODescriptor != null) {
			attrServ = getAttributeConsumingService(spSSODescriptor);
			Extensions ext = spSSODescriptor.getExtensions();
			raExtensions = getRequestedAttributeExtensions(ext);
		}

		LOG.debug("call RelyingParty.newInstance");
		RelyingParty relyingParty = RelyingParty.newInstance(entityId, attrServ, raExtensions);
		return relyingParty;
	}

	/**
	 * Return the EntityDescriptor.
	 * 
	 * @param entityId entity URI string
	 * @return EntityDescriptor object
	 * @throws UApproveException If an exception that was thrown from
	 * MetadataProvider caught.
	 * @see {@link MetadataProvider#getEntityDescriptor(String)}
	 */
	public EntityDescriptor getEntityDescriptor(String entityId)
			throws UApproveException {
		EntityDescriptor entityDescriptor = null;
		try {
			entityDescriptor = metadataProvider.getEntityDescriptor(entityId);
		} catch (MetadataProviderException e) {
			LOG.error("MetadataProviderException", e);
			throw new UApproveException(e);
		}
		LOG.debug("entityDescriptor {}", entityDescriptor);
		return entityDescriptor;
	}

	/**
	 * Return the SPSSODescriptor from EntityDescriptor.
	 * 
	 * <p>Get the SPSSODescriptor element that satisfies the following conditions:</p>
	 * <ul>
	 * <li>At least one of following protocols are supported:
	 * <ul>
	 * <li>urn:oasis:names:tc:SAML:2.0:protocol</li>
	 * <li>urn:oasis:names:tc:SAML:1.1:protocol</li>
	 * <li>urn:oasis:names:tc:SAML:1.0:protocol</li>
	 * </ul>
	 * </li>
	 * <li>Include one or more AttributeConsumingService element</li>
	 * </ul>
	 * 
	 * <p>If an element satisfies the conditions does not found, return null.</p>
	 * 
	 * @param entityDescriptor
	 * @return SPSSODescriptor or null.
	 */
	static SPSSODescriptor getSPSSODescriptor(EntityDescriptor entityDescriptor) {
		for (String protocol: PROTOCOLS) {
			SPSSODescriptor spSSODescriptor = entityDescriptor.getSPSSODescriptor(protocol);
			LOG.debug("SPSSODescriptor for protocol {} is {}", protocol, spSSODescriptor);

			if (spSSODescriptor == null) {
				continue;
			}

			if (spSSODescriptor.getAttributeConsumingServices().size() > 0) {
				return spSSODescriptor;
			}
		}
		return null;
	}
	
	/**
	 * get AttributeConsumingService element from SPSSODescriptor element.
	 * 
	 * get the AttributeConsumingService element with explicit isDefault="true", 
	 * or the first AttributeConsumingService element.
	 */
	private AttributeConsumingService getAttributeConsumingService(
			SPSSODescriptor spSSODescriptor) {
		AttributeConsumingService result = null;

		AttributeConsumingService attrServ = spSSODescriptor.getDefaultAttributeConsumingService();
		if (attrServ != null) {
			LOG.debug("DefaultAttributeConsumingService found");
			result = attrServ;
		} else {
			List<AttributeConsumingService> list = spSSODescriptor.getAttributeConsumingServices();
			if (list != null && !list.isEmpty()) {
				LOG.debug("AttributeConsumingService with index 0 found");
				result = list.get(0);
			} else {
				LOG.debug("SPSSODescriptor has no AttributeConsumingService.");
			}
		}

		return result;
	}

	/**
	 * create RequestedAttributeExtension list from
	 * /EntitiesDescriptor/EntityDescriptor/SPSSODescriptor/Extensions element.
	 * 
	 * RequestedAttributeExtension element belongs to the namespace
	 * "http://www.gakunin.jp/ns/uapprove-jp/metadata".
	 */
	private static List<RequestedAttributeExtension> getRequestedAttributeExtensions(Extensions ext) {
		final RequestedAttributeExtensionBuilder builder1 = new RequestedAttributeExtensionBuilder();
		final DescriptionBuilder builder2 = new DescriptionBuilder();

		List<RequestedAttributeExtension> raExtensionList = new ArrayList<RequestedAttributeExtension>();

		/*
		 * get <RequestedAttributeExtension> element from <Extensions> element.
		 */
		for (XMLObject obj : ext.getUnknownXMLObjects(RequestedAttributeExtension.DEFAULT_ELEMENT_NAME)) {
			LOG.debug("RequestedAttributeExtension:" + obj.getElementQName());

			if (obj instanceof XSAny) {
				XSAny any = (XSAny) obj;

				RequestedAttributeExtension raExtension = builder1.buildObject();				

				AttributeMap attrMap = any.getUnknownAttributes();
				if (LOG.isDebugEnabled()) {
					debugAttributeMap(attrMap);
				}

				if (attrMap.containsKey(FRIENDORY_NAME)) {
					String value = attrMap.get(FRIENDORY_NAME);
					raExtension.setFriendlyName(value);
				}

				/*
				 * get <Description> element from <RequestedAttributeExtension> element.
				 */
				for (XMLObject child : any.getUnknownXMLObjects(Description.DEFAULT_ELEMENT_NAME)) {
					LOG.debug("element name=" + child.getElementQName().toString());

					if (child instanceof XSAny) {
						XSAny any2 = (XSAny) child;

						Description description = builder2.buildObject();

						attrMap = any2.getUnknownAttributes();
						if (LOG.isDebugEnabled()) {
							debugAttributeMap(attrMap);
						}

						if (attrMap.containsKey(XML_LANG)) {
							LocalizedString ls = new LocalizedString(any2.getTextContent(), attrMap.get(XML_LANG));
							description.setDescription(ls);
							LOG.debug("Text={}:{}", ls.getLanguage(), ls.getLocalString());
						}
						// default lang = en?
						description.setParent(raExtension);
						raExtension.getDescriptions().add(description);
					}
				}
				raExtensionList.add(raExtension);
			}
		}
		return raExtensionList;
	}

	private static void debugAttributeMap(AttributeMap attrMap) {
		LOG.debug("attributes=" + attrMap.size());
		for (Map.Entry<QName, String> entry : attrMap.entrySet()) {
			LOG.debug("QName=" + entry.getKey().toString());
			LOG.debug("Value=" + entry.getValue());
		}
	}

}
