package ch.SWITCH.aai.uApprove.idpplugin;

import java.io.IOException;
import java.util.Collection;
import java.util.Date;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.SWITCH.aai.uApprove.components.Attribute;
import ch.SWITCH.aai.uApprove.components.ConfigurationManager;
import ch.SWITCH.aai.uApprove.components.TermsOfUseManager;
import ch.SWITCH.aai.uApprove.components.UApproveException;
import ch.SWITCH.aai.uApprove.idpplugin.StateAndActionEvaluator.Action;
import ch.SWITCH.aai.uApprove.idpplugin.UApproveContextBuilder.UApproveContext;
import ch.SWITCH.aai.uApprove.storage.LogInfo;
import ch.SWITCH.aai.uApprove.storage.UserLogInfo;
import edu.internet2.middleware.shibboleth.common.profile.AbstractErrorHandler;
import edu.internet2.middleware.shibboleth.common.util.HttpHelper;
import edu.internet2.middleware.shibboleth.idp.authn.LoginContext;
import edu.internet2.middleware.shibboleth.idp.session.Session;
import edu.internet2.middleware.shibboleth.idp.util.HttpServletHelper;

/**
 * Class Plugin:
 *
 * Copyright (c) 2011, National Insutitute of Informatics
 * Copyright (c) 2005-2006 SWITCH - The Swiss Education & Research Network
 */
public class Plugin implements Filter {

  private static final String INITPAR_CONFIG = "Config";
  private static Logger logger = LoggerFactory.getLogger(Plugin.class);

  private boolean useTerms = false;

  private Dispatcher dispatcher;
  private StateAndActionEvaluator evaluator;
  private UApproveContextBuilder contextBuilder;
  
  private String authnContextClassRef;

  
  public void init(FilterConfig filterConfig) throws ServletException {
	  // initialize configuration
      ConfigurationManager.initialize(filterConfig.getInitParameter(INITPAR_CONFIG));
      
      // initialize dispatcher
      dispatcher = new Dispatcher(filterConfig.getServletContext());
      
      // initialize state and action evaluator
      evaluator = new StateAndActionEvaluator(dispatcher);

      // initialize state and action evaluator
      contextBuilder = new UApproveContextBuilder(dispatcher);
      
      // initialize terms of use
      String terms = ConfigurationManager.getParam(ConfigurationManager.COMMON_TERMS);
      if (terms != null && !terms.equals("")) {
        try {
          TermsOfUseManager.initalize(terms);
          getStorage();
        } catch (Exception e) {
          throw new UApproveException(e);
        }
        useTerms = true;
        logger.debug("Terms of use version {} loaded", TermsOfUseManager.getVersion());
      } else {
    	logger.debug("No terms of use are configured");
      }

      // initialize sp blacklist
      SPBlacklistManager.initialize(ConfigurationManager.getParam(ConfigurationManager.PLUGIN_SP_BLACKLIST));

      // initialize attribute list
      AttributeList.initialize(ConfigurationManager.getParam(ConfigurationManager.PLUGIN_ATTRIBUTELIST));
    
      // check if the filter only should work on a specified authnContextClassRef
      String accr = filterConfig.getInitParameter("authnContextClassRef");
      if (accr != null && !accr.equals("")) {
        authnContextClassRef = accr;
        logger.debug("uApprove is only enabled on authnContextClassRef {}", authnContextClassRef);
      }
      
      logger.info("uApprove plugin initialized");
  }

  public void destroy() {}

  public void doFilter(ServletRequest servletRequest,
	ServletResponse servletResponse, FilterChain filterChain) throws ServletException, IOException {
      
	logger.trace("BEGIN:");
	HttpServletRequest request;
	HttpServletResponse response;
	try {
	  request = (HttpServletRequest) servletRequest;
	  response = (HttpServletResponse) servletResponse;
	} catch (ClassCastException e) {
       logger.warn("uApprove only works on HTTP requests / responses");
       filterChain.doFilter(servletRequest, servletResponse);
       logger.trace("AFTER(1):");
       return;
    }
	
    try {
	  Action action = evaluator.evaluateAction(request, authnContextClassRef);
	  logger.debug("Action evaluated is {}", action);

	  switch (action) {
	  	case PASS_TO_IDP:
	  		String requestPath = HttpHelper.getRequestUriWithoutContext(request);
	  		logger.debug("RequestPath: {}", requestPath);

	  		dispatcher.dispatchToIdP(request, response, filterChain);

	  		logger.debug("       after {}", requestPath);

	  		if (requestPath.equals("/AuthnEngine")) {
	  			Session session = HttpServletHelper.getUserSession(request);
	  			if (session != null) {
	  				logger.debug("SessionID: {}", session.getSessionID());

	  				/*
	  				 * Authentication is successful, and has been able to
	  				 * retrieve the session.
	  				 * Here clean up attributes release apploval if
	  				 * CheckAlways is set.
	  				 */

	  				LoginContext loginContext = dispatcher.getLoginContext(request);
	  				String principalName = loginContext.getPrincipalName();
	  				if (principalName != null) {
	  					String entityId = loginContext.getRelyingPartyId();
	  					logger.debug("cleanup attributes: {} {}",
	  						principalName, entityId);
	  					LogInfo storage = getStorage();
	  					storage.cleanupAttrsApproval(principalName, entityId);
	  				} else {
	  					logger.debug("principalName is null");
	  				}
	  			} else {
	  				logger.debug("Session is null");
	  			}
	  		}

	  		logger.trace("AFTER(2):");
	  		return;
	  
	  	case CHECK_ACCESS:
	  		checkAccess(request, response, filterChain);
	        logger.trace("AFTER(3):");
		  	return;
	  		
	  }
    } catch (UApproveException e) {
      logger.error("uApprove error: {}", e.getMessage());
      request.setAttribute(AbstractErrorHandler.ERROR_KEY, e);
    }
    logger.trace("AFTER(4):");
  }
  
  
  private void checkAccess(HttpServletRequest request, HttpServletResponse response, FilterChain filterChain) throws UApproveException {
	  
	UApproveContext context = contextBuilder.buildContext(request);
	logger.info("uApprove access: {}", context);
	
	UserLogInfo userInfo = getUserInfo(context);
	logger.debug("Stored user information {}", userInfo);
	
    // set global ARA flag
	boolean globalConsentGiven = ConfigurationManager.makeBoolean(userInfo.getGlobal());

    logger.trace("*** (1) ***");
    logger.trace("check monitoring mode only");
 	if (ConfigurationManager.makeBoolean(ConfigurationManager.getParam(ConfigurationManager.PLUGIN_MONITORING_ONLY))) {
 		logger.info("monitoring mode only");
 		logAccess(context, globalConsentGiven, false);
 		dispatcher.dispatchToIdP(request, response, filterChain);
        return;
	}
 	
    String entityId = context.getRelyingParty().getEntityId();

    logger.trace("*** (2) ***");
    logger.trace("check if relying party is in black list");
	if (SPBlacklistManager.containsItem(entityId)) {
 		logger.info("relying party is in black list");
 		logAccess(context, globalConsentGiven, false);
 		dispatcher.dispatchToIdP(request, response, filterChain);
 		return;
    }

    logger.trace("*** (3) ***");
    logger.trace("check if the user wants to reset the settings");
    if (context.resetConsent()) {
	  logger.info("user wants to reset the settings");
	  dispatcher.dispatchToViewer(request, response, context);
      return;
    }

    logger.trace("*** (4) ***");
    logger.trace("check if the terms of use changed");
    if (useTerms) {
      String termsVersion = TermsOfUseManager.getVersion();
      String userTermsVersion = userInfo.getTermsVersion();
      logger.info("*** (4-1) ***");
      if (!userTermsVersion.equalsIgnoreCase(termsVersion)) {
    	logger.info("user has not agreed to terms of use version {}", termsVersion);

        String attributesSelected = userInfo.getAttributesForProviderId(entityId);
        context.setMandatoryAttributes(getMandatoryAttributeIDs(request, context, attributesSelected));
        if (attributesSelected == null) {
            attributesSelected = "";
        }
        context.setAttributesSelected(attributesSelected.replaceAll("=[^:]*", ""));

    	dispatcher.dispatchToViewer(request, response, context);
        return;
      }
    }
    
    logger.trace("*** (5) ***");
    logger.trace("check if user has given global approval");
    if (globalConsentGiven) {
	  logger.info("user has given global approval");
      logAccess(context, globalConsentGiven, false);
      dispatcher.dispatchToIdP(request, response, filterChain);
	  return;
	}

    logger.trace("*** (6) ***");
    Collection<Attribute> attributes = context.getAttributesReleased();
    if (attributes.isEmpty()) {
  	  logger.info("No attributes will be released");
      logAccess(context, globalConsentGiven, false);
      dispatcher.dispatchToIdP(request, response, filterChain);
      return;
    }
    
    logger.trace("*** (7) ***");
    String attributeVals = Attribute.serializeAttributeValues(context.getAttributesReleased());
    String attributesSelected = userInfo.getAttributesForProviderId(entityId);
    logger.debug("attributeVals=" + attributeVals);
    logger.debug("attributesSelected=" + attributesSelected);
    logger.trace("check if the there are any attributes, which not are already approved for release to relying party");
    if (!Attribute.compareAttributeRelease(attributesSelected, attributeVals)) {
      logger.trace("*** (7-1) ***");
      logger.info("attributes, which not are already approved for release to relying party {}", entityId);
      context.setMandatoryAttributes(getMandatoryAttributeIDs(request, context, attributesSelected));
      if (attributesSelected == null) {
        attributesSelected = "";
      }
      context.setAttributesSelected(attributesSelected.replaceAll("=[^:]*", ""));
      dispatcher.dispatchToViewer(request, response, context);
      return;
    }
    
    logger.trace("*** (8) ***");
    logger.info("user has already given attribute release approval to this relying party {}", entityId);
    logAccess(context, globalConsentGiven, false);
    dispatcher.dispatchToIdP(request, response, filterChain);

    logger.trace("*** (9) ***");
    
    return;
  }
  
  private LogInfo getStorage() throws UApproveException {
	  String storeType = ConfigurationManager.getParam(ConfigurationManager.COMMON_STORE_TYPE);
	  LogInfo.initialize(storeType);
	  return LogInfo.getInstance(); 
  }
  
  private UserLogInfo getUserInfo(UApproveContext context)
		  throws UApproveException {
	LogInfo storage = getStorage();	  
    // getUserInfo
    UserLogInfo userInfo = storage.getData(context.getPrincipal());   
	// create user, if not existent
	if (userInfo == null) {
		logger.debug("user is not existent, create a new user");
		userInfo = storage.addUserLogInfoData(context.getPrincipal(), "1.0", new Date().toString(), "", "no", context.getRelyingParty().getEntityId(), null);
		storage.update(userInfo);
	}
	else { logger.info("user is exists."); }
	logger.info("UserLogInfo:" + userInfo.toString());
	
    // Workaround to get right terms of use
    UserLogInfo userInfoWithRightToU = storage.getDataSimple(context.getPrincipal());
    if (userInfo!= null && userInfoWithRightToU != null) {
  	  userInfo.setTermsVersion(userInfoWithRightToU.getTermsVersion());
    }
	return userInfo;
  }

  private String getMandatoryAttributeIDs(HttpServletRequest request, UApproveContext context, String oldAttributes) throws UApproveException {
  	if (logger.isTraceEnabled()) {
  		logger.trace("oldAttributes=" + oldAttributes);
  	}
  	LogInfo storage = getStorage();
    // getUserInfo
    UserLogInfo userInfo = storage.getData(context.getPrincipal());
    // create user, if not existent
    if (userInfo == null) {
        userInfo = storage.addUserLogInfoData(context.getPrincipal(), "1.0", new Date().toString(), "", "no", context.getRelyingParty().getEntityId(), null);
        storage.update(userInfo);
    }

    String entityId = context.getRelyingParty().getEntityId();
    userInfo.addProviderId(entityId, "");
    storage.update(userInfo, entityId, null);
    String mandatoryAttrIds = Attribute.serializeAttributeIDs(contextBuilder.buildContext(request).getAttributesReleased());
    logger.debug("Mandatory attributes were {}", mandatoryAttrIds);

    if (oldAttributes == null) {
      logger.trace("clearRelease(" + entityId + ")");
      userInfo.clearRelease(entityId);
    } else {
      logger.trace("addProviderId(" + entityId + ")");
      userInfo.addProviderId(entityId, oldAttributes);
    }
    storage.update(userInfo, entityId, null);

    return mandatoryAttrIds;
  }
  
  private void logAccess(UApproveContext context, boolean globalARA, boolean approvalGiven) throws UApproveException {		
	// Is provider access logging enabled
    if (ConfigurationManager.makeBoolean(ConfigurationManager.getParam(ConfigurationManager.PLUGIN_LOG_PROVIDER_ACCESS))) {
    	logger.debug("log provider access {}", context);
    	LogInfo storage = getStorage();
    	if (approvalGiven) {
    		logger.trace("log provider access with approval");
            storage.updateProviderAccess(context.getPrincipal(), context.getRelyingParty().getEntityId(), globalARA);
    	} else {
    		logger.trace("log provider access without approval");
            storage.updateProviderAccessWithNoARA(context.getPrincipal(), context.getRelyingParty().getEntityId());
    	}
    }
		
  }

}
