package ch.SWITCH.aai.uApprove.idpplugin;

import java.util.Collection;
import java.util.HashSet;

import javax.servlet.http.HttpServletRequest;

import org.opensaml.common.xml.SAMLConstants;
import org.opensaml.saml2.metadata.EntityDescriptor;
import org.opensaml.saml2.metadata.SPSSODescriptor;
import org.opensaml.saml2.metadata.provider.MetadataProvider;
import org.opensaml.saml2.metadata.provider.MetadataProviderException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.SWITCH.aai.uApprove.components.Attribute;
import ch.SWITCH.aai.uApprove.components.RelyingParty;
import ch.SWITCH.aai.uApprove.components.UApproveException;
import edu.internet2.middleware.shibboleth.common.attribute.provider.SAML1AttributeAuthority;
import edu.internet2.middleware.shibboleth.common.attribute.provider.SAML2AttributeAuthority;
import edu.internet2.middleware.shibboleth.common.profile.provider.BaseSAMLProfileRequestContext;
import edu.internet2.middleware.shibboleth.common.relyingparty.RelyingPartyConfiguration;
import edu.internet2.middleware.shibboleth.common.relyingparty.provider.SAMLMDRelyingPartyConfigurationManager;
import edu.internet2.middleware.shibboleth.common.session.Session;
import edu.internet2.middleware.shibboleth.idp.authn.LoginContext;

public class UApproveContextBuilder {

	private final Logger logger = LoggerFactory.getLogger(UApproveContextBuilder.class);
	private final Dispatcher dispatcher;
	private final MetadataAccess metadataAccess;

	private static final String SAML1_AA_ID = "shibboleth.SAML1AttributeAuthority";
	private static final String SAML2_AA_ID = "shibboleth.SAML2AttributeAuthority";
	private static final String RP_CM_ID = "shibboleth.RelyingPartyConfigurationManager";

	public final static String RESET_CONSENT_PARAMETER = "_resetconsent";
	private SAMLMDRelyingPartyConfigurationManager relyingPartyConfigurationManager;
	private MetadataProvider metadataProvider;

	UApproveContextBuilder(Dispatcher dispatcher) throws UApproveException {

		this.dispatcher = dispatcher;

		SAML1AttributeAuthority saml1AA = (SAML1AttributeAuthority) dispatcher.getServletContext().getAttribute(SAML1_AA_ID);
		if (saml1AA == null)
			throw new UApproveException("Unable to load SAML1 AA from shibboleth idp context");

		SAML2AttributeAuthority saml2AA = (SAML2AttributeAuthority) dispatcher.getServletContext().getAttribute(SAML2_AA_ID);
		if (saml2AA == null)
			throw new UApproveException("Unable to load SAML2 AA from shibboleth idp context");
          
		relyingPartyConfigurationManager = (SAMLMDRelyingPartyConfigurationManager) dispatcher.getServletContext().getAttribute(RP_CM_ID);
		if (relyingPartyConfigurationManager == null)
			throw new UApproveException("Unable to load RelyingPartyConfigurationManager from shibboleth idp context");

		metadataProvider = relyingPartyConfigurationManager.getMetadataProvider();
		metadataAccess = new MetadataAccess(metadataProvider);

		// initialize AttributeDumper
		AttributeDumper.initialize(saml1AA, saml2AA);       
	}
	
	public UApproveContext buildContext(HttpServletRequest request) throws UApproveException {
		LoginContext loginContext = dispatcher.getLoginContext(request);

		String principal = dispatcher.getPrincipalName(request);

		if (principal == null || "".equals(principal))
			throw new UApproveException("No principal found, assure authentication");
		  
		String relyingPartyId = loginContext.getRelyingPartyId();
		EntityDescriptor entityDescriptor = metadataAccess.getEntityDescriptor(relyingPartyId);

		// get the attributes released for user and relying party;
		final Session session = dispatcher.getSession(request);
		// build request context
		@SuppressWarnings("rawtypes")
			BaseSAMLProfileRequestContext requestCtx = buildRequestContext(principal, session, entityDescriptor);

		Collection<Attribute> attributesReleased = AttributeDumper.getAttributes(requestCtx);

		attributesReleased = AttributeList.sortAttributes(attributesReleased);

		Collection<Attribute> blacklisted = new HashSet<Attribute>();
		for (Attribute attribute: attributesReleased) {
			if (AttributeList.isBlackListed(attribute.attributeID)) {
				blacklisted.add(attribute);
			}
		}
		String attributesBlacklisted = Attribute.serializeAttributeIDs(blacklisted);
		logger.debug("serialized blacklisted attribute:" + attributesBlacklisted);
	      
		boolean resetConsent =  loginContext.getProperty(RESET_CONSENT_PARAMETER) != null
			&& (Boolean) loginContext.getProperty(RESET_CONSENT_PARAMETER);

		RelyingParty relyingParty = metadataAccess.getRelyingPartyInfo(relyingPartyId, entityDescriptor);
		return new UApproveContext(principal, relyingParty, attributesReleased,
				resetConsent, attributesBlacklisted);
	}
	
	/**
	 * Return request context.
	 * 
	 * @param username principal name.
	 * @param session current session.
	 * @param peerMetadata SP entity metadata.
	 * @return 
	 * @throws UApproveException
	 */
	@SuppressWarnings("rawtypes")
	private BaseSAMLProfileRequestContext buildRequestContext(String username,
			Session session, EntityDescriptor peerMetadata) throws UApproveException {
		String spEntityId = peerMetadata.getEntityID();
		RelyingPartyConfiguration relyingPartyConfiguration = relyingPartyConfigurationManager.getRelyingPartyConfiguration(spEntityId);

		String idpEntityId = relyingPartyConfiguration.getProviderId();

		BaseSAMLProfileRequestContext requestCtx = new BaseSAMLProfileRequestContext();
		requestCtx.setRelyingPartyConfiguration(relyingPartyConfiguration);
		requestCtx.setInboundMessageIssuer(spEntityId);
		requestCtx.setOutboundMessageIssuer(idpEntityId);
		requestCtx.setPrincipalName(username);
		requestCtx.setUserSession(session);
		requestCtx.setLocalEntityId(idpEntityId);
		requestCtx.setPeerEntityId(spEntityId);
		requestCtx.setMetadataProvider(metadataProvider);

		try {
			EntityDescriptor localMetadata = metadataProvider.getEntityDescriptor(idpEntityId);
			requestCtx.setLocalEntityMetadata(localMetadata);
			if (localMetadata != null) {
				requestCtx.setLocalEntityRoleMetadata(localMetadata.getIDPSSODescriptor(SAMLConstants.SAML20P_NS));
			} else {
				requestCtx.setLocalEntityRoleMetadata(null);
			}
	
			requestCtx.setPeerEntityMetadata(peerMetadata);
			SPSSODescriptor spSSODescriptor = MetadataAccess.getSPSSODescriptor(peerMetadata);
			requestCtx.setPeerEntityRoleMetadata(spSSODescriptor);
		} catch (MetadataProviderException e) {
			throw new UApproveException(e);
		}
		return requestCtx;
	}

	public class UApproveContext {

		private final String principal;
		private final RelyingParty relyingParty;
		private final Collection<Attribute> attributesReleased;
		private final boolean resetConsent;
		private String mandatoryAttributes = "";
		private String attributesSelected = "";
		private String attributesBlacklisted = "";

		private UApproveContext(String principal, RelyingParty relyingParty,
				Collection<Attribute> attributesReleased, boolean resetConsent,
				String attributesBlacklisted) {
			this.principal = principal;
			this.relyingParty = relyingParty;
			this.attributesReleased = attributesReleased;
			this.resetConsent = resetConsent;
			this.attributesBlacklisted = attributesBlacklisted;
		}

		public String getPrincipal() {
			return principal;
		}

		public RelyingParty getRelyingParty() {
			return relyingParty;
		}

		public Collection<Attribute> getAttributesReleased() {
			return attributesReleased;
		}

		public String getMandatoryAttributes() {
			return mandatoryAttributes;
		}

		public String getAttributesSelected() {
			return attributesSelected;
		}

		public String getAttributesBlacklisted() {
			return attributesBlacklisted;
		}

		public boolean resetConsent() {
			return resetConsent;
		}

		public String toString() {
			return principal+" --> "+relyingParty.getEntityId();
		}

		public void setMandatoryAttributes(String mandatoryAttributes) {
			this.mandatoryAttributes = mandatoryAttributes;
		}

		public void setAttributesSelected(String attributesSelected) {
			this.attributesSelected = attributesSelected;
		}
	}
	
}
