/*
 * Copyright (c) 2010-2012, National Insutitute of Informatics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jp.gakunin.shibboleth.common.attribute.filtering.provider.match.uapprove;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.joda.time.DateTime;
import org.opensaml.common.SAMLObject;
import org.opensaml.saml2.core.AuthnRequest;
import org.opensaml.saml2.metadata.AttributeConsumingService;
import org.opensaml.saml2.metadata.RequestedAttribute;
import org.opensaml.saml2.metadata.RoleDescriptor;
import org.opensaml.saml2.metadata.SPSSODescriptor;
import org.opensaml.xml.XMLObject;
import org.opensaml.xml.schema.XSAny;
import org.opensaml.xml.schema.XSBase64Binary;
import org.opensaml.xml.schema.XSBoolean;
import org.opensaml.xml.schema.XSDateTime;
import org.opensaml.xml.schema.XSInteger;
import org.opensaml.xml.schema.XSString;
import org.opensaml.xml.schema.XSURI;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.SWITCH.aai.uApprove.components.Attribute;
import ch.SWITCH.aai.uApprove.components.ConfigurationManager;
import ch.SWITCH.aai.uApprove.storage.LogInfo;
import edu.internet2.middleware.shibboleth.common.attribute.BaseAttribute;
import edu.internet2.middleware.shibboleth.common.attribute.encoding.AttributeEncoder;
import edu.internet2.middleware.shibboleth.common.attribute.encoding.SAML1AttributeEncoder;
import edu.internet2.middleware.shibboleth.common.attribute.encoding.SAML2AttributeEncoder;
import edu.internet2.middleware.shibboleth.common.attribute.filtering.provider.FilterProcessingException;
import edu.internet2.middleware.shibboleth.common.attribute.filtering.provider.ShibbolethFilteringContext;
import edu.internet2.middleware.shibboleth.common.attribute.filtering.provider.match.basic.AbstractMatchFunctor;
import edu.internet2.middleware.shibboleth.common.attribute.provider.BasicAttribute;
import edu.internet2.middleware.shibboleth.common.profile.provider.SAMLProfileRequestContext;

/**
 * A match function that evaluates an attribute's value against the given match string.
 */
public class AttributeUapproveMatchFunctor extends AbstractMatchFunctor {

    /** Class logger. */
    private final Logger log = LoggerFactory.getLogger(AttributeUapproveMatchFunctor.class);
    private final String APPROVED_ATTR = "ApprovedAttributes";

    /** Whether the attribute must be approved. */
    private boolean isApproved;

    /** Whether the attribute must be requested. */
    private boolean requestedOnly;

    /**
     * Gets whether the attribute must be approved.
     * 
     * @return whether the attribute must be approved
     */
    public boolean isApproved() {
        return isApproved;
    }

    /**
     * Gets whether the attribute must be requested.
     * 
     * @return whether the attribute must be requested
     */
    public boolean isRequestedOnly() {
        return requestedOnly;
    }

    /**
     * Sets whether the attribute must be approved.
     * 
     * @param Approved whether the attribute must be approved
     */
    public void setIsApproved(boolean Approved) {
        isApproved = Approved;
    }

    /**
     * Sets whether the attribute must be requested.
     * 
     * @param requestedOnly whether the attribute must be requested
     */
    public void setRequestedOnly(boolean requestedOnly) {
        this.requestedOnly = requestedOnly;
    }

    /**
     * Evaluates to true if any value for the specified attribute matches the given match string.
     * 
     * {@inheritDoc}
     */

    protected boolean doEvaluatePolicyRequirement(ShibbolethFilteringContext filterContext)
            throws FilterProcessingException {
        log.debug("uApprove policy requirement");

        @SuppressWarnings("rawtypes")
		SAMLProfileRequestContext attributeRequestContext = filterContext.getAttributeRequestContext();

        RoleDescriptor role = attributeRequestContext.getPeerEntityRoleMetadata();
        if (role == null || !(role instanceof SPSSODescriptor)) {
           log.debug("attribute requester's metadata role does not contain attribute requirements");
           return false;
        }

        SPSSODescriptor spSSODescriptor = (SPSSODescriptor) role;

        return getAttributeConsumingService(spSSODescriptor) != null;
    }

    /**
     * Evaluates to true if the given attribute value matches the given match string.
     * 
     * {@inheritDoc}
     */
    protected boolean doEvaluateValue(ShibbolethFilteringContext filterContext, String id, Object attributeValue)
            throws FilterProcessingException {

        int rc = checkRequestedAttribute(filterContext, id, attributeValue);
        if (rc > 0) {
            if (isApproved) {
                return true;
            } else {
                return false;
            }
        } else if (requestedOnly && rc < 0) {
            if (isApproved) {
                return false;
            } else {
                return true;
            }
        }

        String approvedAttrs;
        Map<String, BaseAttribute> attributes = filterContext.getAttributeRequestContext().getAttributes();
        if (attributes != null && attributes.containsKey(APPROVED_ATTR)) {
            log.debug("getAttributes: " + attributes);
            approvedAttrs = (String)attributes.get(APPROVED_ATTR).getValues().iterator().next();
        } else {
            try {
                String entityId = filterContext.getAttributeRequestContext().getPeerEntityMetadata().getEntityID();

                String storeType = ConfigurationManager
                .getParam(ConfigurationManager.COMMON_STORE_TYPE);
                LogInfo.initialize(storeType);
                LogInfo storage = LogInfo.getInstance();
                if(storage == null){
                    log.error("uApprove storage is null");
                    throw new Exception("uApprove storage is null");
                }

                approvedAttrs = storage.getUserReleaseAttrs(filterContext.getAttributeRequestContext().getPrincipalName(), entityId, false);
                log.debug("uApprove release attributes: " + approvedAttrs);
                if (approvedAttrs != null) {
                  approvedAttrs = approvedAttrs.replaceAll("=[^:]*", "");
                }
                BasicAttribute<Object> attribute = new BasicAttribute<Object>();
                attribute.setId(APPROVED_ATTR);
                attribute.getValues().add(approvedAttrs);
                if (attributes == null) {
                    attributes = new HashMap<String, BaseAttribute>();
                }
                attributes.put(APPROVED_ATTR, attribute);
                filterContext.getAttributeRequestContext().setAttributes(attributes);
            } catch (Exception e) {
                log.debug("Unable to get uApprove storage: {}", e.getMessage());
                throw new FilterProcessingException("Unable to get uApprove storage", e);
            }
        }
log.debug("Approved attributes: {}, {}", approvedAttrs, id);
        if (approvedAttrs == null) {
            log.debug("uApprove approve attributes is null " + id);
            if (isApproved) {
                return true;
            } else {
                return false;
            }
        }

        if (Attribute.compareAttributeRelease(approvedAttrs, id)) {
            if (isApproved) {
                return true;
            } else {
                return false;
            }
        } else {
            if (isApproved) {
                return false;
            } else {
                return true;
            }
        }
    }

    private int checkRequestedAttribute(ShibbolethFilteringContext filterContext, String id, Object attributeValue) {
        // Check for SP role.
        RoleDescriptor role = filterContext.getAttributeRequestContext().getPeerEntityRoleMetadata();
        if (!(role instanceof SPSSODescriptor)) {
           log.debug("attribute requester's metadata role does not contain attribute requirements");
           return -1;
        }

        // If the request is an AuthnRequest, check for an AttributeConsumingServiceIndex.
        AttributeConsumingService service = null;
        SAMLObject inbound = filterContext.getAttributeRequestContext().getInboundSAMLMessage();
        if (inbound != null && inbound instanceof AuthnRequest) {
            Integer i = ((AuthnRequest)inbound).getAttributeConsumingServiceIndex();
            if (i != null) {
                List<AttributeConsumingService> services = ((SPSSODescriptor)role).getAttributeConsumingServices();
                for (AttributeConsumingService s : services) {
                    if (s.getIndex() == i) {
                        service = s;
                        break;
                    }
                }
                if (service == null) {
                    log.warn("incoming AuthnRequest's AttributeConsumingServiceIndex did not match the peer's metadata");
                    return -1;
                }
            }
        }

        if (service == null)
            service = getAttributeConsumingService((SPSSODescriptor)role);
        if (service == null) {
            log.debug("the peer's metadata did not contain an AttributeConsumingService descriptor");
            return -1;
        }

        log.debug("using AttributeConsumingService descriptor with index {}", service.getIndex());

        BaseAttribute attr = filterContext.getUnfilteredAttributes().get(id);
        if (attr != null && attr.getValues() != null) {
            List encoders = attr.getEncoders();
            Iterator i = encoders.iterator();
            while (i.hasNext()) {
                AttributeEncoder encoder = (AttributeEncoder) i.next();
                RequestedAttribute requested = findInMetadata(service, encoder);
                if (requested != null) {
                    if (requested.isRequired()) {
                        return 1;
                    }
                    log.debug("found attribute {} requested in metadata", id);
                    List<XMLObject> vals = requested.getAttributeValues();
                    if (vals != null && !vals.isEmpty()) {
                        String v = attributeValue.toString();
                        boolean match = false;
                        for (XMLObject xmlObj : vals) {
                            if (match(xmlObj, v)) {
                                match = true;
                                break;
                            }
                        }
                        if (match == false) {
                            log.debug("attribute {} found in metadata, but value not among those requested", id);
                            return -1;
                        }
                    }
                    return 0;
                }
            }
            log.debug("attribute {} not found in metadata", id);
            return -1;
        }

        return -1;
    }

    private static AttributeConsumingService getAttributeConsumingService(SPSSODescriptor spSSODescriptor) {
        if (spSSODescriptor != null) {
            AttributeConsumingService defaultAttrService = spSSODescriptor.getDefaultAttributeConsumingService();
            if (defaultAttrService != null) {
                return defaultAttrService;
            }

            for (AttributeConsumingService attrService: spSSODescriptor.getAttributeConsumingServices()) {
                return attrService;
            }
        }

        return null;
    }

    private RequestedAttribute findInMetadata(AttributeConsumingService service, AttributeEncoder encoder) {
        List<RequestedAttribute> requested = service.getRequestAttributes();
        for (RequestedAttribute attr : requested) {
            if (attr.getName().equals(encoder.getAttributeName())) {
                String qualifier = null;
                if (encoder instanceof SAML2AttributeEncoder) {
                    qualifier = ((SAML2AttributeEncoder)encoder).getNameFormat();
                }
                else if (encoder instanceof SAML1AttributeEncoder) {
                    qualifier = ((SAML1AttributeEncoder)encoder).getNamespace();
                }
                String nameFormat = attr.getNameFormat();
                if (qualifier == null || nameFormat == null || nameFormat.equals(org.opensaml.saml2.core.Attribute.UNSPECIFIED) || qualifier.equals(nameFormat)) {
                    return attr;
                }
            }
        }
        return null;
    }

    private boolean match(XMLObject xmlObj, String attributeValue) {
        // This is a substitute for a decoder layer that can generate
        // internal comparable value objects out of AttributeValue elements.
        // Short of that, some kind of pluggable comparison object with
        // knowledge of the XML syntax and the internal attribute objects
        // would be needed.
        String toMatch = null;
        if (xmlObj instanceof XSString)
            toMatch = ((XSString)xmlObj).getValue();
        else if (xmlObj instanceof XSURI)
            toMatch = ((XSURI)xmlObj).getValue();
        else if (xmlObj instanceof XSBoolean)
            toMatch = ((XSBoolean)xmlObj).getValue().getValue() ? "1" : "0";
        else if (xmlObj instanceof XSInteger)
            toMatch = ((XSInteger)xmlObj).getValue().toString();
        else if (xmlObj instanceof XSDateTime) {
            DateTime dt = ((XSDateTime)xmlObj).getValue();
            if (dt != null)
                toMatch = ((XSDateTime)xmlObj).getDateTimeFormatter().print(dt);
        }
        else if (xmlObj instanceof XSBase64Binary)
            toMatch = ((XSBase64Binary)xmlObj).getValue();
        else if (xmlObj instanceof XSAny) {
            XSAny wc = (XSAny)xmlObj;
            if (wc.getUnknownAttributes().isEmpty() && wc.getUnknownXMLObjects().isEmpty())
                toMatch = wc.getTextContent();
        }
        if (toMatch != null)
            return toMatch.equals(attributeValue);
        log.warn("unrecognized XMLObject type, unable to match as a string to candidate value");
        return false;
    }
}
