package jp.gakunin.shibboleth.idp.profile.saml1;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;

import org.opensaml.saml1.core.AttributeStatement;
import org.opensaml.saml1.core.NameIdentifier;
import org.opensaml.saml1.core.Response;
import org.opensaml.saml1.core.Statement;
import org.opensaml.saml1.core.StatusCode;
import org.opensaml.ws.transport.http.HTTPInTransport;
import org.opensaml.ws.transport.http.HTTPOutTransport;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.SWITCH.aai.uApprove.components.Attribute;
import ch.SWITCH.aai.uApprove.components.ConfigurationManager;
import ch.SWITCH.aai.uApprove.components.ConsentMode;
import ch.SWITCH.aai.uApprove.components.UApproveException;
import ch.SWITCH.aai.uApprove.idpplugin.AttributeDumper;
import ch.SWITCH.aai.uApprove.storage.LogInfo;
import ch.SWITCH.aai.uApprove.storage.UserLogInfo;

import edu.internet2.middleware.shibboleth.common.attribute.BaseAttribute;
import edu.internet2.middleware.shibboleth.common.profile.ProfileException;
import edu.internet2.middleware.shibboleth.idp.profile.saml1.AttributeQueryProfileHandler;
import edu.internet2.middleware.shibboleth.idp.session.AuthenticationMethodInformation;
import edu.internet2.middleware.shibboleth.idp.session.Session;

/**
 * SAML 1 Attribute Query profile handler.
 */
public class AttributeQueryUApproveProfileHandler extends AttributeQueryProfileHandler {

    /** Class logger. */
    private final Logger log = LoggerFactory.getLogger(AttributeQueryUApproveProfileHandler.class);

    /** URI for transient name format. */
    public static final String TRANSIENT = "urn:mace:shibboleth:1.0:nameIdentifier";

    /** Constructor. */
    public AttributeQueryUApproveProfileHandler() {
        super();

    }

    /** {@inheritDoc} */
    @Override
    public void processRequest(HTTPInTransport inTransport, HTTPOutTransport outTransport) throws ProfileException {
        AttributeQueryUApproveContext requestContext = new AttributeQueryUApproveContext();
        Response samlResponse;

        try {
            decodeRequest(requestContext, inTransport, outTransport);

            if (requestContext.getProfileConfiguration() == null) {
                log.error("SAML 1 Attribute Query profile is not configured for relying party "
                        + requestContext.getInboundMessageIssuer());
                requestContext.setFailureStatus(buildStatus(StatusCode.RESPONDER, StatusCode.REQUEST_DENIED,
                        "SAML 1 Attribute Query profile is not configured for relying party "
                                + requestContext.getInboundMessageIssuer()));
                samlResponse = buildErrorResponse(requestContext);
            } else {
                resolvePrincipal(requestContext);
                String principalName = requestContext.getPrincipalName();

                Session idpSession = getSessionManager().getSession(principalName);
                if (idpSession != null) {
                    requestContext.setUserSession(idpSession);
                    AuthenticationMethodInformation authnInfo = idpSession.getAuthenticationMethods().get(
                            requestContext.getInboundMessageIssuer());
                    if (authnInfo != null) {
                        requestContext.setPrincipalAuthenticationMethod(authnInfo.getAuthenticationMethod());
                    }
                }

                NameIdentifier nameID = requestContext.getSubjectNameIdentifier();
                if (nameID == null) {
                    requestContext.setFailureStatus(buildStatus(StatusCode.RESPONDER, StatusCode.REQUEST_DENIED,
                            "Cannot get NameIdentifier"));
                    String msg = "Cannot get NameIdentifier from SAML request from relying party.";
                    log.warn(msg);
                    throw new ProfileException(msg);
                }
                String format = nameID.getFormat();
                log.debug("NameIdentifier.getFormat()={}", format);

                ArrayList<Statement> statements = new ArrayList<Statement>();
                if (format.equals(TRANSIENT)) {
                    log.trace("match transient name identifier");
                    resolveAttributes(requestContext);
                    createAttributeStatements(statements, requestContext);

                } else {
                    log.trace("not match transient name identifier");
                    createAttributeStatementsBasedOnUserConsent(statements, requestContext, principalName);
                }

                samlResponse = buildResponse(requestContext, statements);
            }
        } catch (ProfileException e) {
            samlResponse = buildErrorResponse(requestContext);
        } catch (UApproveException e) {
            requestContext.setFailureStatus(buildStatus(StatusCode.RESPONDER, null, "Loginfo access failed."));
            String msg = "Error get attributes for principal " + requestContext.getPrincipalName();
            log.error(msg, e);
            samlResponse = buildErrorResponse(requestContext);
        }

        requestContext.setOutboundSAMLMessage(samlResponse);
        requestContext.setOutboundSAMLMessageId(samlResponse.getID());
        requestContext.setOutboundSAMLMessageIssueInstant(samlResponse.getIssueInstant());

        encodeResponse(requestContext);
        writeAuditLogEntry(requestContext);
    }

    private void createAttributeStatementsBasedOnUserConsent(ArrayList<Statement> statements,
            AttributeQueryUApproveContext requestContext, String principalName) throws UApproveException, ProfileException {
        if (log.isDebugEnabled()) {
            log.debug("resolved principal name: {}", requestContext.getPrincipalName());
        }

        if (principalName == null) {
            // May return null when the attribute query name id has been revoked.
            log.error("Could not resolve the name id to principal name. Cannot get the consent mode and all attributes value.");
            return;
        }

        LogInfo storage = getStorage();
        String relyingPartyId = requestContext.getInboundMessageIssuer();

        ConsentMode consentMode = storage.getConsentMode(principalName, relyingPartyId);
        if (log.isDebugEnabled()) {
            log.debug("ConsentMode: {}", consentMode.toString());
        }

        if (consentMode == ConsentMode.CHECK_ALWAYS) {
            log.info("{}'s current consent mode for {} is \"check always\".", principalName, relyingPartyId);
            return;
        }

        resolveAttributes(requestContext);

        if (consentMode == ConsentMode.GLOBAL) {
            log.info("{}'s current consent mode for {} is \"global\".", principalName, relyingPartyId);
        } else {
            UserLogInfo userInfo = storage.getData(principalName);
            String approvedAttrs = userInfo.getAttributesForProviderId(relyingPartyId);
            if (log.isDebugEnabled()) {
                log.debug("attributes from uApprove storage: \"{}\"", approvedAttrs);
            }

            Collection<Attribute> attributes = AttributeDumper.getAttributes(requestContext);
            String currentAttrs = Attribute.serializeAttributeValues(attributes);
            if (log.isDebugEnabled()) {
                log.debug("attributes from DataConnector: \"{}\"", currentAttrs);
            }

            if (Attribute.compareAttributeRelease(approvedAttrs, currentAttrs) == false) {
                log.info("Did not match the attributes from DataConnector against the attributes from uApprove storage.");

                Collection<String> differenceAttributeIds = Attribute.getDifferentAttributeIds(approvedAttrs, currentAttrs);
                log.trace("difference attributes between uApprove storage and DataConnector: \"{}\"", differenceAttributeIds);
                requestContext.removeAttributes(differenceAttributeIds);
            }
        }
        createAttributeStatements(statements, requestContext);

    }

    private void createAttributeStatements(ArrayList<Statement> statements, AttributeQueryUApproveContext requestContext)
            throws ProfileException {
        AttributeStatement attributeStatement = buildAttributeStatement(requestContext,
                "urn:oasis:names:tc:SAML:1.0:cm:sender-vouches");
        if (attributeStatement != null) {
            requestContext.setReleasedAttributes(requestContext.getAttributes().keySet());
            statements.add(attributeStatement);
        }
    }

    private LogInfo getStorage() throws UApproveException {
        String storeType = ConfigurationManager.getParam(ConfigurationManager.COMMON_STORE_TYPE);
        LogInfo.initialize(storeType);
        return LogInfo.getInstance(); 
    }

    /** Basic data structure used to accumulate information as a request is being processed. */
    protected class AttributeQueryUApproveContext extends AttributeQueryContext {

        private void removeAttributes(Collection<String> attributeIds) {
            log.trace("attributes before removing: \"{}\"", getAttributes());

            Map<String, BaseAttribute> attributes = getAttributes();

            for (String attrId: attributeIds) {
                attributes.remove(attrId);
            }
            setAttributes(attributes);

            log.trace("attributes after removing: \"{}\"", getAttributes());
        }
    }
}