package ch.SWITCH.aai.uApprove.viewer;

import java.io.IOException;
import java.util.Collection;
import java.util.Date;
import java.util.Locale;
import java.util.Map;
import java.util.List;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.SWITCH.aai.uApprove.components.Attribute;
import ch.SWITCH.aai.uApprove.components.ConfigurationManager;
import ch.SWITCH.aai.uApprove.components.Crypt;
import ch.SWITCH.aai.uApprove.components.ConsentMode;
import ch.SWITCH.aai.uApprove.components.RelyingParty;
import ch.SWITCH.aai.uApprove.components.RequestedAttributeDescription;
import ch.SWITCH.aai.uApprove.components.TermsOfUseManager;
import ch.SWITCH.aai.uApprove.components.UApproveException;
import ch.SWITCH.aai.uApprove.storage.LogInfo;
import ch.SWITCH.aai.uApprove.storage.UserLogInfo;
import ch.qos.logback.classic.LoggerContext;
import ch.qos.logback.classic.joran.JoranConfigurator;
import ch.qos.logback.core.joran.spi.JoranException;

/**
 * uApprove Controller Servlet
 * 
 * Copyright (c) 2011, National Insutitute of Informatics
 * Copyright (c) 2005-2006 SWITCH - The Swiss Education & Research Network
 * 
 * Purpose: Steering class which decides what button the user pressed and then
 * updates/stores the relevant information.
 * 
 * @author: C.Witzig, based on the original code of F.Poroli 8.12.2005
 */
public class Controller extends HttpServlet {

	private static final long serialVersionUID = 1L;

	public static final String INITPAR_CONFIG = "Config";

	public static final String PAGE_TERMS = "/terms-of-use.jsp";
	public static final String PAGE_TERMS_DECLINED = "/terms-of-use-declined.jsp";
	public static final String PAGE_ATTRIBUTES = "/attributes.jsp";
	public static final String PAGE_ATTRIBUTES_DECLINED = "/attributes-declined.jsp";
	public static final String PAGE_ATTRIBUTES_CHECK = "/attributes-check.jsp";
	public static final String PAGE_RESET = "/reset-approvals.jsp";
	public static final String PAGE_ERROR = "/error.jsp";

	public static final String GETPAR_TERMS_CONFIRM = "terms-confirm";
	public static final String GETPAR_TERMS_AGREE = "terms-agree";
	public static final String GETPAR_TERMS_DECLINE = "terms-decline";
	public static final String GETPAR_TERMS_DECLINE_BACK = "terms-decline-back";
	public static final String GETPAR_ATTRIBUTES_NEXT = "attributes-next";
	public static final String GETPAR_ATTRIBUTES_CANCEL = "attributes-cancel";
	public static final String GETPAR_ATTRIBUTES_SEND = "attributes-send";
	public static final String GETPAR_ATTRIBUTES_CONSENT_MODE = "attributes-consent-mode";
	public static final String GETPAR_ATTRIBUTES_BACK = "attributes-back";
	public static final String GETPAR_ATTRIBUTES_DECLINE_BACK = "attributes-decline-back";
	public static final String GETPAR_RELEASED_ATTRIBUTES = "approved-attributes[]";

	public static final String GETPAR_STANDALONE = "standalone-next-url";
	public static final String GETPAR_RESET_CANCEL = "reset-cancel";
	public static final String GETPAR_RESET_CONFIRM = "reset-confirm";

	public static final String RB_TERMS = "terms-of-use";
	public static final String RB_TERMS_DECLINED = "terms-of-use-declined";
	public static final String RB_ATTRIBUTES = "attributes";
	public static final String RB_ATTRIBUTES_CHECK = "attributes-check";
	public static final String RB_ATTRIBUTES_DECLINED = "attributes-declined";
	public static final String RB_RESET = "reset-approvals";

	public static final String SESKEY_PRINCIPAL = "principal";
	public static final String SESKEY_RETURNURL = "returnURL";
	public static final String SESKEY_RELYINGPARTY = "relyingParty";
	public static final String SESKEY_ATTRIBUTES = "attributes";
	public static final String SESKEY_GLOBAL_CONSENT_POSSIBLE = "globalConsentPossible";
	public static final String SESKEY_ATTRIBUTES_CONSENT_MODE = "attributesConsentMode";
	public static final String SESKEY_LOCALE = "locale";
	public static final String SESKEY_ERROR = "error";
	public static final String SESKEY_MANDATORY_ATTRIBUTES = "mandatoryAttributes";
	public static final String SESKEY_SELECTED_ATTRIBUTES = "selectedAttributes";
	public static final String SESKEY_BLACKLISTED_ATTRIBUTES = "blacklistedAttributes";
	
	private static Logger LOG = LoggerFactory.getLogger(Controller.class);

	private LogInfo storage;
	private boolean useTerms = false;
	private Crypt crypt;

	public void init() throws ServletException {
		try {
			// config init
			ConfigurationManager.initialize(getServletContext().getInitParameter(INITPAR_CONFIG));

			// initializing logback
			LoggerContext lc = (LoggerContext) LoggerFactory.getILoggerFactory();
			try {
				JoranConfigurator configurator = new JoranConfigurator();
				configurator.setContext(lc);
				lc.reset();
				configurator.doConfigure(ConfigurationManager.getParam(ConfigurationManager.VIEWER_LOGBACK_CONFIG));
			} catch (JoranException je) {
				throw new UApproveException(je);
			}

			// initialize terms
			String terms = ConfigurationManager.getParam(ConfigurationManager.COMMON_TERMS);
			if (terms != null && !terms.equals("")) {
				try {
					TermsOfUseManager.initalize(terms);
				} catch (Exception e) {
					throw new UApproveException(e);
				}
				useTerms = true;
				LOG.debug("Terms of use loaded, version {}", TermsOfUseManager.getVersion());
			} else {
				LOG.debug("No Terms of use are used");
			}

			// storage init
			String storeType = ConfigurationManager.getParam(ConfigurationManager.COMMON_STORE_TYPE);
			LogInfo.initialize(storeType);
			storage = LogInfo.getInstance();
			LOG.debug("LogInfo (storage) initialized with mode {}", storeType);
		} catch (UApproveException e) {
			LOG.error("uApprove servlet initialization failed", e);
		}

		// initialize secret for decryption
		String sharedSecret = ConfigurationManager.getParam(ConfigurationManager.COMMON_SHARED_SECRET);
		crypt = new Crypt(sharedSecret);
	}

	protected void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
		try {

			// initialization
			HttpSession session = request.getSession();
			
			// get principal
			String principal = crypt.decrypt(request.getParameter(ConfigurationManager.HTTP_PARAM_PRINCIPAL));
			LOG.debug("principal=" + principal);
			session.setAttribute(SESKEY_PRINCIPAL, principal);

			// get returnURL
			String returnURL = request.getParameter(ConfigurationManager.HTTP_PARAM_RETURNURL);
			LOG.debug("returnURL=" + returnURL);
			session.setAttribute(SESKEY_RETURNURL, returnURL);

			// get relying party
			String relyingPartySerialized = crypt.decrypt(request.getParameter(ConfigurationManager.HTTP_PARAM_RELYINGPARTY));
			LOG.debug("RP decrypted, serialized: {}", relyingPartySerialized);
			RelyingParty relyingParty = RelyingParty.newInstance(relyingPartySerialized);

			String entityId = relyingParty.getEntityId();
			LOG.debug("entityId=" + entityId);
			session.setAttribute(SESKEY_RELYINGPARTY, relyingParty);

			if (request.getParameter(ConfigurationManager.HTTP_PARAM_RESET) != null) {
				// start edit flow
				LOG.info("user want to edit the attribute release approval");
				getServletContext().getRequestDispatcher(PAGE_RESET).forward(request, response);
				return;
			} else {
				// Start flow
				
				// get released attributes
				String serializedAttributesReleased = crypt.decrypt(request.getParameter(ConfigurationManager.HTTP_PARAM_ATTRIBUTES));
				LOG.debug("serializedAttributesReleased are {}", serializedAttributesReleased);
				Collection<Attribute> attributesReleased = Attribute.unserializeAttributes(serializedAttributesReleased);
				session.setAttribute(SESKEY_ATTRIBUTES, attributesReleased);

				// get globalConsentPossible
				boolean globalConsentPossible = ConfigurationManager.makeBoolean(ConfigurationManager.getParam(ConfigurationManager.VIEWER_GLOBAL_CONSENT));
				LOG.debug("globalConsentPossible is {}", globalConsentPossible);
				session.setAttribute(SESKEY_GLOBAL_CONSENT_POSSIBLE, globalConsentPossible);

				// get the locale
				Locale locale = createLocale(request.getLocale(), ConfigurationManager.getParam(ConfigurationManager.VIEWER_USELOCALE));
				LOG.debug("locale is {}", locale);
				session.setAttribute(SESKEY_LOCALE, locale);

				// get mandatory attributes
				String mandatoryAttributes = crypt.decrypt(request
					.getParameter(ConfigurationManager.HTTP_PARAM_MANDATORY_ATTRIBUTES));
				LOG.debug("mandatoryAttributes is {}" ,mandatoryAttributes);
				session.setAttribute(SESKEY_MANDATORY_ATTRIBUTES, mandatoryAttributes);

				// get selected attributes
				String attributesSelected = crypt.decrypt(request
					.getParameter(ConfigurationManager.HTTP_PARAM_SELECTED_ATTRIBUTES));
				LOG.debug("attributesSelected is {}", attributesSelected);
				session.setAttribute(SESKEY_SELECTED_ATTRIBUTES, attributesSelected);

				// get blacklisted attributes
 				String attributesBlacklisted = crypt.decrypt(request
 					.getParameter(ConfigurationManager.HTTP_PARAM_BLACKLISTED_ATTRIBUTES));
				LOG.debug("attributesBlacklisted is {}", attributesBlacklisted);
 				session.setAttribute(SESKEY_BLACKLISTED_ATTRIBUTES, attributesBlacklisted);

				// getUserInfo
				UserLogInfo userInfo = storage.getData(principal);
				LOG.debug("userInfo is {}", userInfo);

				// set default consent mode
				session.setAttribute(Controller.SESKEY_ATTRIBUTES_CONSENT_MODE,
						ConsentMode.CHECK_ALWAYS);

				if (userInfo == null) {
					if (useTerms) {
						LOG.debug("first visit of the user, redirect to terms page");
						getServletContext().getRequestDispatcher(PAGE_TERMS).forward(request, response);
					} else {
						LOG.debug("first visit of the user, redirect to attributes page");
						getServletContext().getRequestDispatcher(PAGE_ATTRIBUTES).forward(request, response);
					}
					return;
				}

				boolean globalConsentGiven = ConfigurationManager.makeBoolean(userInfo.getGlobal());

				if (globalConsentGiven) {
					session.setAttribute(Controller.SESKEY_ATTRIBUTES_CONSENT_MODE,
							ConsentMode.GLOBAL);
				} else if (userInfo.containsProviderId(entityId) &&
						!userInfo.checkAlways(entityId)) {
					session.setAttribute(Controller.SESKEY_ATTRIBUTES_CONSENT_MODE,
							ConsentMode.THE_SERVICE);
				}

				// has user agreed to the current terms version
				if (useTerms && !userInfo.getTermsVersion().equals(TermsOfUseManager.getVersion())) {
					LOG.info("current terms version are not agreed by user, redirect to the terms page");
					getServletContext().getRequestDispatcher(PAGE_TERMS).forward(request, response);
					return;
				}

				LOG.info("Terms are not used or current terms version are agreed by user, redirect to the attributes page");
				getServletContext().getRequestDispatcher(PAGE_ATTRIBUTES).forward(request, response);
				return;
			}
		} catch (UApproveException e) {
			doError(request, response, e);
		}
	}

	protected void doGet(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {
		try {
			// initialization
			HttpSession session = request.getSession();
			ConfigurationManager.initialize(getServletContext().getInitParameter(INITPAR_CONFIG));

			String principal = (String) session.getAttribute(SESKEY_PRINCIPAL);
			LOG.debug("principal is {}", principal);
			
			if (principal == null || principal.equals("")) {
				throw new UApproveException("Principal is not set");
			}

			RelyingParty relyingParty = (RelyingParty) session.getAttribute(SESKEY_RELYINGPARTY);
			LOG.debug("entityId is {}", relyingParty.getEntityId());

			UserLogInfo userInfo = storage.getData(principal);
			// Ugly hack
			UserLogInfo userInfoWithRightToU = storage.getDataSimple(principal);
			if (userInfo != null && userInfoWithRightToU != null) {
				userInfo.setTermsVersion(userInfoWithRightToU.getTermsVersion());
			}

			LOG.debug("userInfo is {}", userInfo);

			@SuppressWarnings("unchecked")
			Collection<Attribute> attributesReleased = (Collection<Attribute>) session.getAttribute(SESKEY_ATTRIBUTES);
			LOG.debug("attributesReleased are {}", attributesReleased);

			String[] selectedParameterValues = request.getParameterValues(GETPAR_RELEASED_ATTRIBUTES);
			// TODO if selectedParameters is null?

			String returnURL = (String) session.getAttribute(SESKEY_RETURNURL);
			LOG.debug("returnURL is {}", returnURL);

			if (isGetParSet(request, GETPAR_TERMS_CONFIRM)) {
				LOG.debug("coming from terms confirmed");
				// check if the user agreed the terms
				if (isGetParSet(request, GETPAR_TERMS_AGREE)) {
					LOG.info("user agreed terms, store, redirect to attributes, if necessary");
					storeUserBasic(userInfo, principal, TermsOfUseManager.getVersion());
					userInfo = storage.getData(principal);

					// check if user has to be redirected to the attributes
					// approval page
					if (ConfigurationManager.makeBoolean(userInfo.getGlobal())) {
						LOG.info("user has global consent");
						response.sendRedirect(response.encodeRedirectURL(returnURL));
						return;
					}

					String approvedAttributes = userInfo.getAttributesForProviderId(relyingParty.getEntityId());
					LOG.debug("approvedAttributes are {}", approvedAttributes);
					
					if (attributesReleased.isEmpty()) {
						LOG.info("No attributes are released for this relying party");
						response.sendRedirect(response.encodeRedirectURL(returnURL));
						return;
					}
					
					String serializedAttributeIDs = Attribute.serializeAttributeIDs(attributesReleased);
					LOG.debug("serializedAttributeIDs to be released are {}", serializedAttributeIDs);
					if (Attribute.compareAttributeRelease(approvedAttributes, serializedAttributeIDs)) {
						LOG.info("user has already approved attributes for this relying party");
						response.sendRedirect(response.encodeRedirectURL(returnURL));
						return;
					}

					getServletContext().getRequestDispatcher(PAGE_ATTRIBUTES).forward(request, response);
					return;

				} else {
					LOG.info("user dont agreed terms, redirect again to terms");
					getServletContext().getRequestDispatcher(PAGE_TERMS).forward(request, response);
					return;
				}
			}

			if (isGetParSet(request, GETPAR_TERMS_DECLINE)) {
				LOG.debug("coming from terms declined, redirect to decline page");
				getServletContext().getRequestDispatcher(PAGE_TERMS_DECLINED).forward(request, response);
				return;
			}

			if (isGetParSet(request, GETPAR_TERMS_DECLINE_BACK)) {
				LOG.debug("coming from terms declined back, redirect to terms page");
				getServletContext().getRequestDispatcher(PAGE_TERMS).forward(request, response);
				return;
			}

			// "Next" from attributes.jsp
			if (isGetParSet(request, GETPAR_ATTRIBUTES_NEXT)) {
				LOG.debug("coming from attributes Next");

				String consentModeStr = request.getParameter(GETPAR_ATTRIBUTES_CONSENT_MODE);
				LOG.info("consent mode=" + consentModeStr);
				// TODO if consentMode is null ...
				ConsentMode cm = ConsentMode.valueOf(consentModeStr);
				session.setAttribute(SESKEY_ATTRIBUTES_CONSENT_MODE, cm);

				StringBuilder attributesSelected = new StringBuilder(":");
				for (String namval: selectedParameterValues) {
					String[] x = namval.split("=");
					attributesSelected.append(x[0]).append(":");
				}

				LOG.debug("attributesSelected is {}", attributesSelected);
				session.setAttribute(SESKEY_SELECTED_ATTRIBUTES, attributesSelected.toString());

				getServletContext().getRequestDispatcher(PAGE_ATTRIBUTES_CHECK)
						.forward(request, response);
				return;
			}

			// "Cancel" from attributes.jsp
			if (isGetParSet(request, GETPAR_ATTRIBUTES_CANCEL)) {
				LOG.debug("coming from attributes cancel");
				getServletContext().getRequestDispatcher(PAGE_ATTRIBUTES_DECLINED).forward(request, response);
				return;
			}

			// "Send" from attributes-check.jsp
			if (isGetParSet(request, GETPAR_ATTRIBUTES_SEND)) {
				LOG.info(
						"user gave attribute release consent, store, redirect to returnURL {}",
						response.encodeRedirectURL(returnURL));

				String termsVersion = useTerms ? TermsOfUseManager.getVersion() : "";
				String attributes = getAttributesReleased(relyingParty.getEntityId(), selectedParameterValues);
				ConsentMode consentMode = (ConsentMode) session.getAttribute(SESKEY_ATTRIBUTES_CONSENT_MODE);
				LOG.info("consent mode=" + consentMode);
				// TODO if consentMode is null ...
				String global = consentMode.equals(ConsentMode.GLOBAL) ? "yes" : "no";

				LOG.debug("store: principal="
						+ principal
						+ " entityId="
						+ relyingParty.getEntityId()
						+ " attributes="
						+ attributes
						+ " terms="
						+ termsVersion
						+ " globalConsent="
						+ global);
				storeUser(userInfo, principal, termsVersion, relyingParty.getEntityId(), attributes, global, consentMode);
				response.sendRedirect(response.encodeRedirectURL(returnURL));
				return;
			}

			// "Back" from attributes-check.jsp
			// TODO GETPAR_ATTRIBUTES_BACK ?
			if (isGetParSet(request, GETPAR_ATTRIBUTES_BACK)) {
				//LOG.debug("coming from attributes declined, redirect to decline page");
				LOG.debug("coming from attributes-check back, redirect to attributes page");
				getServletContext().getRequestDispatcher(PAGE_ATTRIBUTES).forward(request, response);
				return;
			}

			// "Back" from attributes-declined.jsp
			if (isGetParSet(request, GETPAR_ATTRIBUTES_DECLINE_BACK)) {
				LOG.info("coming from attributes declined back, redirect to attributes page");
				getServletContext().getRequestDispatcher(PAGE_ATTRIBUTES).forward(request, response);
				return;
			}

			// coming here is not ok
			throw new UApproveException("uApprove viewer flow error");
		} catch (UApproveException e) {
			doError(request, response, e);
		}
	}

	private void storeUser(UserLogInfo userInfo, String principal, String termsVersion,
			String entityId, String attributesReleased, String global, ConsentMode cm)
					throws UApproveException {

		LOG.debug("storeUser");
		LogInfo storage = LogInfo.getInstance();
		if (userInfo == null) {
			LOG.debug("create new user");
			userInfo = storage.addUserLogInfoData(principal, "1.0",
					new Date().toString(), termsVersion, global, entityId,
					attributesReleased);
			// TODO
		} else {
			userInfo.setOndate(new Date().toString());
			userInfo.setTermsVersion(termsVersion);
			userInfo.setGlobal(global);

			switch (cm) {
			case CHECK_ALWAYS:
				LOG.debug("add ProviderId: " + entityId);
				userInfo.addProviderId(entityId, attributesReleased);
				if (!userInfo.checkAlways(entityId)) {
					userInfo.addCheckAlways(entityId);
				}
				break;
			case THE_SERVICE:
				LOG.debug("add ProviderId: " + entityId);
				userInfo.addProviderId(entityId, attributesReleased);
				if (userInfo.checkAlways(entityId)) {
					userInfo.clearCheckAlways(entityId);
				}
				break;
			case GLOBAL:	// global
				LOG.debug("all clear.");
				userInfo.clearRelease();
				break;
			}
		}
		storage.update(userInfo, entityId, cm);
	}

	private void storeUserBasic(UserLogInfo userInfo, String principal,
			String termsVersion) throws UApproveException {
		LOG.debug("storeUserBasic");
		LogInfo storage = LogInfo.getInstance();

		if (userInfo == null) {
			LOG.debug("create new user");
			userInfo = storage.addUserLogInfoData(principal, "1.0",
					new Date().toString(), termsVersion, "no", null, null);
		} else {
			userInfo.setTermsVersion(termsVersion);
			userInfo.setOndate(new Date().toString());
		}
		storage.update(userInfo);
	}

	public static boolean isGetParSet(HttpServletRequest request, String par) {
		if (request.getParameter(par) == null || request.getParameter(par).equals("")) {
			return false;
		}
		return true;
	}

	public static Locale createLocale(Locale requestLocale, String enforced) {

		if (enforced != null && !enforced.equals(""))
			return new Locale(enforced.substring(0, enforced.indexOf("_")), enforced.substring(enforced.indexOf("_") + 1));

		if (requestLocale != null)
			return requestLocale;

		return Locale.ENGLISH;
	}

	// parse resource (sp) host
	public static String getResourceHost(String entityId) {
		int i1 = entityId.indexOf("//");
		int i2 = entityId.indexOf("/", i1 + 2);
		LOG.debug("entityId received = \"" + entityId + "\"");

		// return just the sp.example.org component out of
		// https://sp.example.org/shibboleth
		if (i2 >= 0)
			entityId = entityId.substring(i1 + 2, i2);
		else if (i1 >= 0)
			entityId = entityId.substring(i1 + 2);

		// return just the sp.example.org component out of
		// urn:mace:federation.org:sp.example.org
		if (entityId.indexOf(':') >= 0) {
			entityId = entityId.substring(entityId.lastIndexOf(':') + 1);
		}
		LOG.debug("hostname extracted = \"" + entityId + "\"");
		return entityId;
	}

	public static String getRelyingPartyName(RelyingParty relyingParty,
			Locale locale) {

		Map<String, String> names = relyingParty.getRpNames();
		String name = names.get(locale.getLanguage());
		if (name != null)
			return name;
		name = names.get(Locale.ENGLISH.getLanguage());
		if (name != null)
			return name;

		return getResourceHost(relyingParty.getEntityId());
	}

	public static String getRelyingPartyDesc(RelyingParty relyingParty, Locale locale) {
		Map<String, String> descs = relyingParty.getRpDescriptions();
		String desc = descs.get(locale.getLanguage());
		if (desc != null)
			return desc;
		desc = descs.get(Locale.ENGLISH.getLanguage());
		if (desc != null)
			return desc;

		return "";
	}

	public static List<String> getMandatoryAttributes(RelyingParty relyingParty) {
		return relyingParty.getMandatoryAttributes();
	}

	public static Map<String, RequestedAttributeDescription> getReqdAttrDescs(RelyingParty relyingParty) {
		return relyingParty.getReqdAttrDescs();
	}

	// resolves displayName due to locale
	public static String resolveDisplayName(Attribute attribute, Locale locale) {
		String name = attribute.attributeNames.get(locale.getLanguage());
		if (name == null)
			name = attribute.attributeNames.get(Locale.ENGLISH.getLanguage());
		if (name == null)
			name = attribute.attributeID;
		return name;
	}

	// resolves displayDescription due to locale
	public static String resolveDisplayDesc(Attribute attribute, Locale locale) {
		String desc = attribute.attributeDescriptions.get(locale.getLanguage());
		if (desc == null)
			desc = attribute.attributeDescriptions.get(Locale.ENGLISH.getLanguage());
		if (desc == null)
			desc = "";
		return desc;
	}

	public static void doError(HttpServletRequest request, HttpServletResponse response, UApproveException e) throws ServletException, IOException {
		HttpSession session = request.getSession();
		session.setAttribute(SESKEY_ERROR, e);
		LOG.error("uApprove Error", e);
		session.getServletContext().getRequestDispatcher(PAGE_ERROR).forward(request, response);
	}

	public static String getAttributesReleased(String entityId, String[] attributes) throws UApproveException {
		LOG.debug("attributes: " + attributes);
		LogInfo storage = LogInfo.getInstance();

		String attributesReleased = "";

		if (attributes != null) {
			for (String attribute: attributes) {
				LOG.debug("relased attribute: " + attribute);
				attributesReleased = attributesReleased + ":" + attribute;
			}
			attributesReleased += ":";
		}

		LOG.debug("Approved attributes are " + attributesReleased);

		return attributesReleased;
	}
}
